#!/usr/bin/perl

# Author : Robert Vojta (robert.vojta@mandrake.cz) 
#
# Contributors : Damien Krotkine (dams@libconf.net)
#
# Copyright (C) 2004 Robert Vojta (robert.vojta@mandrake.cz), Damien Krotkine (dams@libconf.net)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

package Libconf::System::Desktop;
use strict;
use vars qw(@ISA);
use Libconf qw(:helpers);
use Libconf::System::Desktop::General;
use Libconf::Glueconf::X::Gdm;
use Libconf::Glueconf::X::Xdm;

@ISA = qw(Libconf);

sub new {
    my ($pkg) = @_;
    my $self = bless {
                      general => new Libconf::System::Desktop::General,
                      desktop_managers => {
			     gdm => new Libconf::Glueconf::X::Gdm('/etc/X11/gdm/gdm.conf'),
			     #for now we use gdm to read kdm config. Will be cleaned with new libconf templates
	                     kdm => new Libconf::Glueconf::X::Gdm('/usr/share/config/kdm/kdmrc'),
	                     xdm => new Libconf::Glueconf::X::Xdm('/etc/X11/xdm/xdm-config'),
		            },
	             }, $pkg;
    $self->synchronize_autologin_read();
    $self;
}

sub readConf {
    my ($self) = @_;
    $self->{$_}->readConf foreach qw(general);
    $self->{desktop_managers}->{$_}->readConf foreach qw(gdm kdm xdm);
    $self->synchronize_autologin_read();
}

sub writeConf {
    my ($self) = @_;
    $self->synchronize_autologin_write();
    $self->{$_}->writeConf() foreach qw(general);
    $self->{desktop_managers}->{$_}->writeConf() foreach qw (gdm kdm xdm);
}

# read the info to determine autologin
sub synchronize_autologin_read {
    my ($self) = @_;
    my $autologin;
    my $user;
    if ($self->{general}->{desktop} eq 'KDE') {
        $autologin = text2bool($self->{desktop_managers}->{kdm}->{'X-:0-Core'}->{AutoLoginEnable});
	$user =  $self->{desktop_managers}->{kdm}->{'X-:0-Core'}->{AutoLoginUser};
    } elsif ($self->{general}->{desktop} eq 'GNOME') {
        $autologin = text2bool($self->{desktop_managers}->{gdm}->{daemon}->{AutomaticLoginEnable});
	$user = $self->{desktop_managers}->{gdm}->{daemon}->{AutomaticLogin};
    } else {
	$autologin = to_bool($user = $self->{general}->{autologin});
    }

    $self->set_autologin($autologin, $user);
}

# dispatch autologin info
sub synchronize_autologin_write {
    my ($self) = @_;
    my $autologin = to_bool(my $user = $self->{general}->{autologin});
    $self->set_autologin($autologin, $user);
}

sub set_autologin {
    my ($self, $autologin, $user) = @_;
    $self->{desktop_managers}->{kdm}->{'X-:0-Core'}->{AutoLoginEnable} = bool2truefalse($autologin);
    $self->{desktop_managers}->{kdm}->{'X-:0-Core'}->{AutoLoginUser} = $user;

    $self->{desktop_managers}->{gdm}->{daemon}->{AutomaticLoginEnable} = bool2truefalse($autologin);
    $self->{desktop_managers}->{gdm}->{daemon}->{AutomaticLogin} = $user;

    $self->{general}->{autologin} = $autologin ? $user : undef;
}

1;
